#include "include/page.h"
#include "include/riscv.h"
#include "include/type.h"
#include "include/kmallo.h"
#include "include/assert.h"
#include "include/panic.h"
#include "include/printk.h"
#include "include/string.h"

static uint64_t* bitmap;
static uint64_t bitmap_size;
static void* memory_start;
static uint64_t total_pages;

int init_physical_memory(void* start, uint64_t size) {
    if (!start || size == 0) {
        return PAGE_INIT_ERROR;
    }
    #ifdef DEBUG
    printk("[OS: Memory] Initializing physical memory manager...\n");
    printk("             Memory start: 0x%lx\n", (uint64_t)start);
    printk("             Memory size: %lu bytes\n", size);
    #endif
    total_pages = size / PAGE_SIZE;
    bitmap_size = (total_pages + 63) / 64;
    uint64_t bitmap_bytes = bitmap_size * 8;
    bitmap = (uint64_t*)start;
    memset(bitmap, 0, bitmap_bytes);
    memory_start = (void*)(((uint64_t)start + bitmap_bytes + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1));
    uint64_t bitmap_pages = ((uint64_t)memory_start - (uint64_t)start + PAGE_SIZE - 1) / PAGE_SIZE;
    for (uint64_t i = 0; i < bitmap_pages; i++) {
        bitmap[BITMAP_WORD(i)] |= (1UL << BITMAP_BIT(i));
    }
    return PAGE_INIT_OK;
}

void* alloc_page(void) {
    printk("bitmap: %lx, memory_start: %lx\n", *bitmap, memory_start);
    Assert(bitmap || memory_start, "Physical memory not initialized!");

    for (uint64_t i = 0; i < total_pages; i++) {
        if (!(bitmap[BITMAP_WORD(i)] & (1UL << BITMAP_BIT(i)))) {
            bitmap[BITMAP_WORD(i)] |= (1UL << BITMAP_BIT(i));
            void* page = (void*)((uint64_t)memory_start + i * PAGE_SIZE);
            memset(page, 0, PAGE_SIZE);
            return page;
        }
    }
    panic("Out of memory!");
    return NULL;
}

void free_page(void* page) {
    if (!page || !bitmap || !memory_start) return;
    
    Assert(((uint64_t)page < (uint64_t)memory_start || \
        (uint64_t)page >= (uint64_t)memory_start + total_pages * PAGE_SIZE), \
        "Invalid page address in free_page!");
    
    uint64_t idx = ((uint64_t)page - (uint64_t)memory_start) / PAGE_SIZE;
    bitmap[BITMAP_WORD(idx)] &= ~(1UL << BITMAP_BIT(idx));
}

void set_page_used(void* page) {
    if (!page) return;
    uint64_t idx = ((uint64_t)page - (uint64_t)memory_start) / PAGE_SIZE;
    if (idx >= total_pages) return;
    bitmap[BITMAP_WORD(idx)] |= (1UL << BITMAP_BIT(idx));
}

int is_page_used(void* page) {
    if (!page) return 0;
    uint64_t idx = ((uint64_t)page - (uint64_t)memory_start) / PAGE_SIZE;
    if (idx >= total_pages) return 0;
    return (bitmap[BITMAP_WORD(idx)] & (1UL << BITMAP_BIT(idx))) != 0;
}