#include "include/vm.h"
#include "include/page.h"
#include "include/assert.h"
#include "include/printk.h"

void init_pagetable(pagetable_t* pt) 
{
    for (int i = 0; i < PTE_NUM; i++) {
        pt->entries[i] = 0;
    }
}

pagetable_t* create_pagetable() {
    pagetable_t* pt = (pagetable_t*)alloc_page();
    if (!pt) return NULL;
    init_pagetable(pt);
    return pt;
}

int free_pagetable(pagetable_t* pt, int level) {
    if (level < 0) return -1;
    
    for (int i = 0; i < PTE_NUM; i++) {
        pte_t pte = pt->entries[i];
        if (pte & PTE_V) {
            uint64_t pa = ((pte >> 10) << 12);
            if (level > 0) {
                pagetable_t* child_pt = (pagetable_t*)pa;
                free_pagetable(child_pt, level - 1);
            }
        }
    }
    free_page((void*)pt);
    return 0;
}

inline uint64_t make_pte(uint64_t ppn, uint64_t perm) {
    return (ppn << PTE_PPN_SHIFT) | perm;
}

pte_t* walk_pte(pagetable_t* root, uint64_t va, int alloc) 
{
    if (!root) return NULL;

    pagetable_t* pt = root;
    uint64_t vpn[3] = {VPN_0(va), VPN_1(va), VPN_2(va)};
    
    for (int level = 2; level > 0; level--) {
        pte_t* pte = &pt->entries[vpn[level]];
        if (!(*pte & PTE_V)) {
            if (!alloc) return NULL;
            pagetable_t* new_pt = create_pagetable();
            if (!new_pt) return NULL;
            *pte = make_pte((uint64_t)new_pt >> 12, PTE_V);
        }
        pt = (pagetable_t*)get_pa(*pte);
    }
    return &pt->entries[vpn[0]];
}


int map_page(pagetable_t* root, uint64_t va, uint64_t pa, uint32_t perm) {
    pte_t* pte = walk_pte(root, va, 1);
    if (!pte) return -1;
    if (*pte & PTE_V)
        return -1;
    *pte = make_pte(pa >> 12, perm | PTE_V);
    return 0;
}


uint64_t walk(pagetable_t* root, uint64_t va) {
    pte_t* pte = walk_pte(root, va, 0);
    if (pte == NULL || !(*pte & PTE_V))
        return 0;
    return get_pa(*pte);
}

int unmap_page(pagetable_t* root, uint64_t va) {
    pte_t* pte = walk_pte(root, va, 0);
    if (!pte || !(*pte & PTE_V))
        return -1;
    *pte = 0;
    return 0;
}

pagetable_t* copy_pagetable(pagetable_t* src) {
    pagetable_t* dst = create_pagetable();
    if (!dst) return NULL;

    for (int i = 0; i < PTE_NUM; i++) {
        pte_t pte = src->entries[i];
        if (pte & PTE_V) {
            if ((pte & (PTE_R|PTE_W|PTE_X)) == 0) {
                pagetable_t* child = (pagetable_t*)get_pa(pte);
                pagetable_t* new_child = copy_pagetable(child);
                if (!new_child) {
                    free_pagetable(dst, 2);
                    return NULL;
                }
                dst->entries[i] = make_pte((uint64_t)new_child >> 12, get_perm(pte));
            } else {
                dst->entries[i] = pte;
            }
        }
    }
    return dst;
}

int map_range(pagetable_t* pt, uint64_t va, uint64_t pa, uint64_t size, uint32_t perm) {
    uint64_t a;
    uint64_t last = (va + size - 1) & ~(PAGE_SIZE-1);
    
    for(a = va & ~(PAGE_SIZE-1); a <= last; a += PAGE_SIZE, pa += PAGE_SIZE) {
        if(map_page(pt, a, pa, perm) != 0)
            return -1;
    }
    return 0;
}

int unmap_range(pagetable_t* pt, uint64_t va, uint64_t size) {
    uint64_t a;
    uint64_t last = (va + size - 1) & ~(PAGE_SIZE-1);
    
    for(a = va & ~(PAGE_SIZE-1); a <= last; a += PAGE_SIZE) {
        if(unmap_page(pt, a) != 0)
            return -1;
    }
    return 0;
}

int check_permission(pagetable_t* pt, uint64_t va, uint32_t perm) {
    pte_t* pte = walk_pte(pt, va, 0);
    if (!pte || !(*pte & PTE_V))
        return -1;
    if ((*pte & perm) != perm)
        return -1;
    return 0;
}

